#include <WiFi.h>
#include <PubSubClient.h>
#include <ESPmDNS.h>
#include <ArduinoOTA.h>
#include "time.h"

const char* ssid       = "Talknet";
const char* password   = "yona1952";

const char* ntpServer = "pool.ntp.org";
const long  gmtOffset_sec = 3600;
const int   daylightOffset_sec = 3600;

const PROGMEM char* MQTT_SERVER_IP = "10.0.0.4";
const PROGMEM uint16_t MQTT_SERVER_PORT = 1883;
const PROGMEM char* MQTT_USER = "pi";
const PROGMEM char* MQTT_PASSWORD = "raspberry";

int freq = 38000;
int channel = 0;
int resolution = 8;

int SensorPin = 23; // blue
int GNDPin = 22;    // purple
int VCCPin = 14;    // red
int PWMPin = 32;    // yellow

int diameter = 133; // mm
int sections = 4;
int delta_x = diameter * 3.1415 / sections;
double max_speed = 7.0;
double min_time= delta_x * 3.6 / max_speed; 

int threshold;
int prev_is_black = 0;
int mm_traveled = 0;
int prev_time = millis();
double curr_speed;
int prev_day;
char buffer[256];
struct tm timeinfo;
int count, max_count = 50;
bool is_moving = false;
double lpf = 1.0;

WiFiClient wifiClient;
PubSubClient client(wifiClient);


// ---------------------------------------------------------------------------------------
//                                    S e t u p
// ---------------------------------------------------------------------------------------
void setup() {
  Serial.begin(9600);

  // IO
  pinMode(SensorPin, INPUT);
  pinMode(GNDPin, OUTPUT);
  pinMode(VCCPin, OUTPUT);
  pinMode(PWMPin, OUTPUT);
  digitalWrite(SensorPin, HIGH); // pullup
  digitalWrite(GNDPin, LOW);
  digitalWrite(VCCPin, HIGH);

  pinMode(LED_BUILTIN, OUTPUT);

  ledcSetup(channel, freq, resolution);
  ledcAttachPin(PWMPin, channel);
  ledcWrite(channel, 10);

  Serial.println();
  /*
  Serial.print("Waiting for white reflection...");
  while (digitalRead(SensorPin) == 1)
    ;
  */
  Serial.println("Ready");
  Serial.print("Calibrating threshold : ");
  threshold = 2;
  Serial.println();
  Serial.print("Threshold = ");
  Serial.println(threshold);
  ledcWrite(channel, threshold);

  // Wifi
  Serial.printf("Connecting to %s ", ssid);
  WiFi.begin(ssid, password);
  while (WiFi.status() != WL_CONNECTED) {
      delay(500);
      Serial.print(".");
  }
  Serial.println(" CONNECTED");
  
  //init and get the time
  configTime(gmtOffset_sec, daylightOffset_sec, ntpServer);

  // init the MQTT connection
  client.setServer(MQTT_SERVER_IP, MQTT_SERVER_PORT);

  // mDNS
  if (MDNS.begin("chinchi")) 
    Serial.println("mDNS responder started");
  else
    Serial.println("Error setting up MDNS responder!");
      

  // OTA
  ArduinoOTA.onStart([]() {
  });
  ArduinoOTA.onEnd([]() {
  });
  ArduinoOTA.onProgress([](unsigned int progress, unsigned int total) {
  });
  ArduinoOTA.onError([](ota_error_t error) {
    Serial.printf("Error[%u]: ", error);
    if (error == OTA_AUTH_ERROR) Serial.println("Auth Failed");
    else if (error == OTA_BEGIN_ERROR) Serial.println("Begin Failed");
    else if (error == OTA_CONNECT_ERROR) Serial.println("Connect Failed");
    else if (error == OTA_RECEIVE_ERROR) Serial.println("Receive Failed");
    else if (error == OTA_END_ERROR) Serial.println("End Failed");
  });
  ArduinoOTA.begin();

  //browseService("mqtt", "tcp");

}

// ---------------------------------------------------------------------------------------
void graph()
{
  int m = 10;
  while (true)
  {
     for (int color = 0; color < m; color++)
    { 
      ledcWrite(channel, color);
      delay(10);
      int reading = digitalRead(SensorPin); 
      Serial.print(color);
      Serial.print(" ");  
      Serial.print(reading*m);
      Serial.println(" ");        
    }
  }
}

int find_threshold()
{
  int reading;
  int result[5];
  // test 
  for (int test=0; test<5; test++) {
    for (int color = 0; color < 255; color++)
    { 
      ledcWrite(channel, color);
      delay(10);
      reading = digitalRead(SensorPin); 
      if (reading == 0) {
        Serial.print(color);
        Serial.print(" ");  
        result[test] = color;
        break;
      }
      
    }
  }
  // sort
  for (int i=0; i<4; i++)
    for (int j=i+1; j<5; j++)
      if (result[i] > result[j]) {
        int tmp = result[i];
        result[i] = result[j];
        result[j] = tmp;
      }
  // return median
  return result[2];
}


// ---------------------------------------------------------------------------------------
//                                      W I F I
// ---------------------------------------------------------------------------------------


void reconnect() {
  // Loop until we're reconnected
  digitalWrite(LED_BUILTIN, HIGH); // led off
  while (!client.connected()) {
    Serial.println("Connecting to MQTT...");
    if (client.connect("ESP32Client", MQTT_USER, MQTT_PASSWORD )) {
      Serial.println("connected");
    } else {
      Serial.print("failed with state ");
      Serial.print(client.state());
      for (int seconds=0; seconds<2; seconds++)
      {
        digitalWrite(LED_BUILTIN, HIGH);
        delay(500);
        digitalWrite(LED_BUILTIN, LOW); // led on
        delay(500);
      }
    }
  }
  digitalWrite(LED_BUILTIN, LOW); // led on
}

void browseService(const char * service, const char * proto){
    Serial.printf("Browsing for service _%s._%s.local. ... ", service, proto);
    int n = MDNS.queryService(service, proto);
    if (n == 0) {
        Serial.println("no services found");
    } else {
        Serial.print(n);
        Serial.println(" service(s) found");
        for (int i = 0; i < n; ++i) {
            // Print details for each service found
            Serial.print("  ");
            Serial.print(i + 1);
            Serial.print(": ");
            Serial.print(MDNS.hostname(i));
            Serial.print(" (");
            Serial.print(MDNS.IP(i));
            Serial.print(":");
            Serial.print(MDNS.port(i));
            Serial.println(")");
        }
    }
    Serial.println();
}

// ---------------------------------------------------------------------------------------
//                                     L o o p
// ---------------------------------------------------------------------------------------
void loop() {

  if (!client.connected()) {
    reconnect();
  }

  client.loop();
  ArduinoOTA.handle();

  if(!getLocalTime(&timeinfo)){
    Serial.println("Failed to obtain time");
    return;
  }

  lpf = 0.9 * lpf + 0.1 * digitalRead(SensorPin);
  bool is_black = lpf > 0.5;
  int curr_time = millis();
  int delta_t = curr_time - prev_time;
/*
  buffer[count++] = '0'+is_black;
  if (count >= max_count)
  {
    buffer[count] = '\0';
    client.publish("casamami/hamstermeter/debug", buffer);
    count = 0;
  }
  delay(10);
  return;

*/

  // 11111 1 1.0 
  // 11110 1 0.8
  // 11100 1 0.6
  // 11000 0 0.4
  // 10000 0 0.2
  // 00000 0 0.0
  
  if (is_black != prev_is_black) 
  {
    digitalWrite(LED_BUILTIN, is_black);   
    prev_is_black = is_black;
    
    if (delta_t > min_time) // check this is not just noise (speed > 10 kph)
    {
      mm_traveled += delta_x; // circumfence is 42 cm. so each quadrant is 10.5 cm or 105 mm
      is_moving = true;
      curr_speed = delta_x * 3.6 /  delta_t; // (mm/ms = m/s)  * 3.6 [k/h] 
      dtostrf(delta_t, 0, 0, buffer);
      client.publish("casamami/hamstermeter/dt", buffer);
      dtostrf(mm_traveled/1000.0, 0, 1, buffer);
      client.publish("casamami/hamstermeter/today", buffer);
      dtostrf(curr_speed, 0, 1, buffer);
      client.publish("casamami/hamstermeter/speed", buffer);
    }
    prev_time = curr_time;
  }

  // Timeout
  if ((delta_t > 2000) && (is_moving))
  {
    curr_speed = 0;
    is_moving = false;
    dtostrf(delta_t, 0, 0, buffer);
    client.publish("casamami/hamstermeter/dt", buffer);
    dtostrf(curr_speed, 0, 1, buffer);
    client.publish("casamami/hamstermeter/speed", buffer);
    digitalWrite(LED_BUILTIN, LOW);   
  }

  Serial.print(mm_traveled/1000.0,1);
  Serial.print(" ");
  Serial.print(curr_speed,1);
  Serial.println();

  // day change
  if (timeinfo.tm_mday != prev_day)
  {
    //Serial.print(&timeinfo, "%A, %B %d %Y %H:%M:%S ");
    Serial.print("Day change ");
    Serial.println(timeinfo.tm_mday);
    prev_day = timeinfo.tm_mday;
    dtostrf(mm_traveled/1000.0, 0, 1, buffer);
    client.publish("casamami/hamstermeter/yesturday", buffer);
    mm_traveled = 0;
    dtostrf(mm_traveled/1000.0, 0, 1, buffer);
    client.publish("casamami/hamstermeter/today", buffer);
  }
  
  delay(10);
  
}

